from ..core.combine_text import combine_text

_BULLET_TO_TEXT_SPACING = "   "
_BULLETS = ["●", "○", "⦿"]

def bullet_list(*text: str, base_indentation_level: int = 0):
    """Formats a list of text items as a bullet list with customizable indentation."""
    formatted_lines = []

    for item_text in text:
        stripped_text = item_text.lstrip()
        leading_spaces = len(item_text) - len(stripped_text)
        indentation_level = base_indentation_level + (leading_spaces // 4)

        bullet = _BULLETS[indentation_level % len(_BULLETS)]
        indent_str = "  " * indentation_level

        line = combine_text(indent_str, bullet, _BULLET_TO_TEXT_SPACING, stripped_text)
        formatted_lines.append(line)

    return "\n".join(formatted_lines)



from typing import Dict, Union, Optional, Callable

from .color import color
from .gradient import gradient

def character_color_map(
    char_colors: Dict[str, Union[color, gradient]], 
    default_color: Optional[Union[color, gradient]] = None
) -> Callable[[str], str]:
    """Creates a function that applies character-specific colors to a given text."""
    normalized_colors = {}
    
    for char, colorizer in char_colors.items():
        if not isinstance(colorizer, (color, gradient)):
            continue
            
        if char.isalpha():
            normalized_colors[char.lower()] = colorizer
            normalized_colors[char.upper()] = colorizer
        else:
            normalized_colors[char] = colorizer

    def color_text(text: str) -> str:
        """Applies the character-specific colors to the provided text."""
        if not text:
            return text
        
        lines = text.split('\n')
        total_lines = len(lines)
        
        result = []
        for y, line in enumerate(lines):
            line_result = []
            for x, char in enumerate(line):
                colorizer = normalized_colors.get(char, default_color)
                
                if colorizer is None:
                    line_result.append(char)
                    continue
                    
                if isinstance(colorizer, gradient):
                    color_factor = y / (total_lines - 1) if total_lines > 1 else 0
                    color = colorizer.at(color_factor)
                    line_result.append(color(char))
                else:
                    line_result.append(colorizer(char))
            
            result.append(''.join(line_result))
                
        return '\n'.join(result)
    
    return color_text


from ...colors.character_color_map import character_color_map
from ...core.combine_text import combine_text
from ...colors.presets import presets
from ...styles.bold import bold

from datetime import datetime

_COLOR = presets.blue
_MAP = character_color_map({
    "[": _COLOR,
    "]": _COLOR,
})

def timed_print(*text: str):
    """Prints a message prefixed with the current time (H:M:S) on the left."""
    
    now = datetime.now()
    current_time_str = now.strftime("%H:%M:%S")

    combined_text = combine_text(*text)
    prefix = _MAP(f"[{current_time_str}]")

    print(f"{prefix} {combined_text}")



from ..core.style import style
from ..colors.color import color
from ..enums.layers import layers

def background(color: color, *text, force_ansi: bool = False):
    """Applies a foreground color to the given text."""
    return style(color.to_template(layers.BACKGROUND))(*text, force_ansi=force_ansi)


from ..color import color

def from_html(hex_code: str) -> color:
    """
    Converts a hex color code to a color object.
    Supports both 3-digit and 6-digit hex codes.
    """

    hex_code = hex_code.lstrip('#')
    hex_code = ''.join(c * 2 for c in hex_code) if len(hex_code) == 3 else hex_code
    return color(*(int(hex_code[i:i+2], 16) for i in (0, 2, 4)))


from ...core.combine_text import combine_text
from ...colors.presets import presets
from ...styles.bold import bold

_INFO_COLOR = presets.dodger_blue

def info(*text: str) -> None:
    """Prints an informational message."""

    combined_text = combine_text(*text)
    print(bold(f"{_INFO_COLOR("INFO")} {combined_text}"))



# colors
from .colors.character_color_map import character_color_map
from .colors.color import color
from .colors.gradient import gradient
from .colors.presets import presets

## parser
from .colors.parser.from_hsl import from_hsl
from .colors.parser.from_html import from_html
from .colors.parser.rainbow_text import rainbow_text
from .colors.parser.random_color import random_color

# styles
from .styles.background import background
from .styles.bold import bold
from .styles.bullet_list import bullet_list
from .styles.foreground import foreground
from .styles.italic import italic
from .styles.strikethrough import strikethrough
from .styles.underline import underline

# utils
from .utils.supports_ansi import supports_ansi

## prints
from .utils.prints.debug import debug
from .utils.prints.error import error
from .utils.prints.info import info
from .utils.prints.mutable import mutable_print
from .utils.prints.success import success
from .utils.prints.timed import timed_print
from .utils.prints.warn import warn

__init__ = [
    "character_color_map",
    "color"
    "gradient",
    "parser",

    "from_hsl",
    "from_html",
    "rainbow_text",
    "random_color",

    "bold",
    "italic",
    "underline",
    "strikethrough",
    "bullet_list",

    "supports_ansi",

    "debug",
    "error",
    "info",
    "mutable_print",
    "success",
    "timed_print",
    "warn",
]


from ..color import color

def from_hsl(h: float, s: float, l: float) -> color:
    """Converts HSL color values to an RGB color object."""
    h = h % 360 / 360
    q = l * (1 + s) if l < 0.5 else l + s - l * s
    p = 2 * l - q
    
    def hue_rgb(t: float) -> float:
        """Helper function to convert hue to RGB component."""
        t = t % 1
        if t < 1/6: return p + (q - p) * 6 * t
        if t < 1/2: return q
        if t < 2/3: return p + (q - p) * (2/3 - t) * 6
        return p

    return color(*(int(hue_rgb(h + i) * 255) for i in (1/3, 0, -1/3)))


from ...core.combine_text import combine_text
from ...colors.presets import presets
from ...styles.bold import bold

_WARN_COLOR = presets.bright_orange

def warn(*text: str) -> None:
    """Prints an warning message."""

    combined_text = combine_text(*text)
    print(bold(f"{_WARN_COLOR("WARN")} {combined_text}"))
    


from typing import Dict, Union, Tuple

from .color import color
from ..enums.layers import layers
from ..enums.gradient_type import gradient_type
from ..core.combine_text import combine_text
from .presets import presets


def sqrt(n: float, square: float = 2) -> float:
    return pow(n, (1 / square))

class gradient:
    __slots__ = ('colors', 'start', 'end', 'type')

    def __init__(
        self,
        colors: Union[color, Dict[float, color]],
        start: Tuple[float, float] = (0.5, 0),
        end: Tuple[float, float] = (0.5, 1),
        type: gradient_type = gradient_type.LINEAR
    ):
        """Initializes a gradient with specified colors, start and end positions, and gradient type."""
        self.colors = {0.0: colors, 1.0: colors} if isinstance(colors, color) else colors
        self.start = start
        self.end = end
        self.type = type
        
    def _get_position_factor(self, x: int, y: int, max_x: int, max_y: int) -> float:
        """Calculates the position factor based on the x and y coordinates."""
        if max_x <= 1 and max_y <= 1:
            return 0.0
            
        norm_x = x / (max_x - 1) if max_x > 1 else 0.0
        norm_y = y / (max_y - 1) if max_y > 1 else 0.0
        
        if self.type == gradient_type.LINEAR:
            dx = self.end[0] - self.start[0]
            dy = self.end[1] - self.start[1]
            pdx = norm_x - self.start[0]
            pdy = norm_y - self.start[1]
            
            dot = dx * pdx + dy * pdy
            mag_squared = dx * dx + dy * dy
            
            if mag_squared == 0:
                return 0.0
            
            return max(0.0, min(1.0, dot / mag_squared))
        
        dx = norm_x - self.start[0]
        dy = norm_y - self.start[1]
        
        radius = sqrt((self.end[0] - self.start[0]) ** 2 + (self.end[1] - self.start[1]) ** 2)
        
        if radius == 0:
            return 0.0
            
        return max(0.0, min(1.0, sqrt(dx * dx + dy * dy) / radius))
    
    def at(self, position: float) -> color:
        """Returns the color at a specific position in the gradient."""
        if position in self.colors:
            return self.colors[position]
            
        positions = sorted(self.colors.keys())
        
        if position <= positions[0]:
            return self.colors[positions[0]]
        if position >= positions[-1]:
            return self.colors[positions[-1]]
            
        for i in range(len(positions) - 1):
            start_pos = positions[i]
            end_pos = positions[i + 1]
            
            if start_pos <= position <= end_pos:
                start_color = self.colors[start_pos]
                end_color = self.colors[end_pos]
                
                # Avoid division by zero if start_pos and end_pos are identical
                if end_pos == start_pos:
                    return start_color
                    
                factor = (position - start_pos) / (end_pos - start_pos)
                
                # Linearly interpolate RGB components
                return color(
                    int(start_color.r + (end_color.r - start_color.r) * factor),
                    int(start_color.g + (end_color.g - start_color.g) * factor),
                    int(start_color.b + (end_color.b - start_color.b) * factor)
                )
            
        return self.colors[positions[0]] 
    
    def __call__(self, text: str, layer: layers = layers.TEXT) -> str:
        """Applies the gradient to the given text."""
        if not text:
            return text
            
        if len(text) == 1:
            return self.at(0.0)(text, layer)
            
        lines = text.split('\n')
        max_y = len(lines)
        max_x = 0
        for line in lines:
            if len(line) > max_x:
                max_x = len(line)
        
        styled_lines = []
        for y, line in enumerate(lines):
            styled_line_chars = []
            for x, char in enumerate(line):
                position_factor = self._get_position_factor(x, y, max_x, max_y)
                char_color = self.at(position_factor)
                styled_line_chars.append(char_color(char, layer))
                
            for x_pad in range(len(line), max_x):
                position_factor = self._get_position_factor(x_pad, y, max_x, max_y)
                char_color = self.at(position_factor)
                styled_line_chars.append(char_color(' ', layer))

            styled_lines.append("".join(styled_line_chars))
            
        return "\n".join(styled_lines)



from ...core.combine_text import combine_text
from ...colors.presets import presets
from ...styles.bold import bold

_DEBUG_COLOR = presets.yellow

def debug(*text: str) -> None:
    """Prints a debug message."""

    combined_text = combine_text(*text)
    print(bold(f"{_DEBUG_COLOR("DEBU")} {combined_text}"))
    


from ...core.combine_text import combine_text
from ...enums.gradient_type import gradient_type
from ...enums.layers import layers
from ..gradient import gradient
from ..presets import presets

_GRADIENT_COLOR = gradient(
    {
        0: presets.red,
        1/8: presets.bright_orange,
        2/8: presets.yellow,
        3/8: presets.bright_green,
        4/8: presets.cyan,
        5/8: presets.deep_blue,
        6/8: presets.violet,
        7/8: presets.magenta,
        1: presets.pink,
    },
    start=(0, 0.5),
    end=(1, 0.5),
    type=gradient_type.LINEAR
)

def rainbow_text(*text: str) -> str:
    """Returns a rainbow-colored text string."""
    combined_text = combine_text(*text)
    return _GRADIENT_COLOR(combined_text, layer=layers.TEXT)



from ..utils.supports_ansi import supports_ansi
from .combine_text import combine_text
from dataclasses import dataclass, field

_ANSI_SUPPORTED = supports_ansi()

@dataclass(slots=True, frozen=True)
class style:
    template: str

    _ansi_supported: bool = field(default=_ANSI_SUPPORTED, init=False, repr=False)

    def __call__(self, *text, force_ansi: bool = False) -> str:
        """ Allows the Style instance to be called like a function to format text. """
        return self.format(combine_text(*text), force_ansi)
    
    def format(self, text: str, force_ansi: bool) -> str:
        """
        Formats the given text with the stored template,
        respecting ANSI support settings.
        """
        if not self._ansi_supported and not force_ansi:
            return text
         
        return self.template.format(text)


from ...core.combine_text import combine_text
from ...colors.presets import presets
from ...styles.bold import bold

_ERROR_COLOR = presets.red

def error(*text: str) -> None:
    """Prints an error message."""

    combined_text = combine_text(*text)
    print(bold(f"{_ERROR_COLOR("ERRO")} {combined_text}"))
    


from random import choice, random, randint
from typing import List, Union, Optional

from ..color import color
from ..gradient import gradient

def random_color(colors: Optional[List[Union[color, gradient]]] = None) -> color:
    if not colors or colors == None or len(colors) == 0:
        return color(randint(0, 255), randint(0, 255), randint(0, 255))
            
    chosen_item = choice(colors)

    if isinstance(chosen_item, color):
        return chosen_item
    else:
        return chosen_item.get_color_at(random())


from sys import stdout

def mutable_print(text: str):
    """Prints a message that can be updated in place."""
    
    stdout.write(text)
    stdout.flush()

    def update(new_text: str):
        """Updates the printed message with new text."""

        stdout.write('\r\033[K' + new_text)
        stdout.flush()

    return update



from dataclasses import dataclass
from ..enums.layers import layers

_ANSI_TEMPLATE = {
    layers.TEXT: "\033[38;2;{r};{g};{b}m{{}}\033[39m",
    layers.BACKGROUND: "\033[48;2;{r};{g};{b}m{{}}\033[49m",
    layers.FULL: "\033[38;2;{r};{g};{b};48;2;{r};{g};{b}m{{}}\033[39m\033[49m"
}

@dataclass(slots=True, frozen=True)
class color:
    """Represents a color in RGB format with methods to convert to different formats."""
    r: int
    g: int
    b: int

    def __post_init__(self):
        """Ensure RGB values are within the range [0, 255]"""
        object.__setattr__(self, 'r', max(0, min(255, self.r)))
        object.__setattr__(self, 'g', max(0, min(255, self.g)))
        object.__setattr__(self, 'b', max(0, min(255, self.b)))

    def __str__(self) -> str:
        """Return a string representation of the color in RGB format."""
        return f"RGB({self.r}, {self.g}, {self.b})"
        
    def __repr__(self) -> str:
        """Return a detailed string representation for debugging."""
        return f"color(r={self.r}, g={self.g}, b={self.b})"


    def to_tuple(self) -> tuple[int, int, int]:
        """Return the color as an RGB tuple."""
        return (self.r, self.g, self.b)

    def to_hex(self) -> str:
        """Return the color in hexadecimal format."""
        return f"#{self.r:02x}{self.g:02x}{self.b:02x}"

    def to_template(self, layer: layers = layers.TEXT) -> str:
        """
        Return an ANSI escape code template for the color.
        The template can be used to format text with the specified color.
        """
        if layer not in _ANSI_TEMPLATE:
            raise ValueError(f"Unsupported layer for ANSI template: {layer}")

        return _ANSI_TEMPLATE[layer].format(r=self.r, g=self.g, b=self.b)

    def __call__(self, text: str, layer: layers = layers.TEXT) -> str:
        """Applies the ANSI color code to the given text for the specified layer."""
        template = self.to_template(layer)
        return template.format(text)


from .color import color

class presets:
    # --- Grayscale ---
    white = color(255, 255, 255)
    snow = color(255, 250, 250)
    honey_dew = color(240, 255, 240)
    mint_cream = color(245, 255, 250)
    azure = color(240, 255, 255)
    alice_blue = color(240, 248, 255)
    ghost_white = color(248, 248, 255)
    white_smoke = color(245, 245, 245)
    gainsboro = color(220, 220, 220)
    light_gray = color(211, 211, 211)
    silver = color(192, 192, 192)
    dark_gray = color(169, 169, 169)
    gray = color(128, 128, 128)
    dim_gray = color(105, 105, 105)
    slate_gray = color(112, 128, 144)
    light_slate_gray = color(119, 136, 153)
    dark_slate_gray = color(47, 79, 79)
    black = color(0, 0, 0)
    
    # --- Red Shades ---
    red = color(255, 0, 0)
    dark_red = color(139, 0, 0)
    fire_brick = color(178, 34, 34)
    crimson = color(220, 20, 60)
    indian_red = color(205, 92, 92)
    light_coral = color(240, 128, 128)
    salmon = color(250, 128, 114)
    dark_salmon = color(233, 150, 122)
    light_salmon = color(255, 160, 122)
    tomato = color(255, 99, 71)
    orange_red = color(255, 69, 0)
    brick_red = color(151, 0, 0)
    bright_red = color(196, 40, 28)
    ruby = color(224, 17, 95)
    scarlet = color(255, 36, 0)
    maroon = color(128, 0, 0)
    burgundy = color(128, 0, 32)
    radical_red = color(255, 50, 100)

    # --- Pink Shades ---
    pink = color(255, 192, 203)
    light_pink = color(255, 182, 193)
    hot_pink = color(255, 105, 180)
    deep_pink = color(255, 20, 147)
    pale_violet_red = color(219, 112, 147)
    medium_violet_red = color(199, 21, 133)
    lavender_blush = color(255, 240, 245)
    misty_rose = color(255, 228, 225)
    thistle = color(216, 191, 216)
    fuchsia = color(255, 0, 255)

    # --- Orange Shades ---
    orange = color(255, 165, 0)
    dark_orange = color(255, 140, 0)
    coral = color(255, 127, 80)
    sandy_brown = color(244, 164, 96)
    dark_goldenrod = color(184, 134, 11)
    goldenrod = color(218, 165, 32)
    light_goldenrod_yellow = color(250, 250, 210)
    burly_wood = color(222, 184, 135)
    tan = color(210, 180, 140)
    peach_puff = color(255, 218, 185)
    navajo_white = color(255, 222, 173)
    moccasin = color(255, 228, 181)
    papaya_whip = color(255, 239, 213)
    blanched_almond = color(255, 235, 205)
    antique_white = color(250, 235, 215)
    bright_orange = color(255, 148, 0)
    persimmon = color(236, 88, 0)

    # --- Yellow Shades ---
    yellow = color(255, 255, 0)
    light_yellow = color(255, 255, 224)
    lemon_chiffon = color(255, 250, 205)
    pale_goldenrod = color(238, 232, 170)
    khaki = color(240, 230, 140)
    dark_khaki = color(189, 183, 107)
    gold = color(255, 215, 0)
    olive = color(128, 128, 0)
    yellow_green = color(154, 205, 50)
    bright_yellow = color(245, 205, 48)
    canary_yellow = color(255, 255, 153)

    # --- Green Shades ---
    green = color(0, 128, 0)
    lime = color(0, 255, 0)
    dark_green = color(0, 100, 0)
    forest_green = color(34, 139, 34)
    sea_green = color(46, 139, 87)
    medium_sea_green = color(60, 179, 113)
    light_sea_green = color(32, 178, 170)
    pale_green = color(152, 251, 152)
    light_green = color(144, 238, 144)
    medium_spring_green = color(0, 250, 154)
    spring_green = color(0, 255, 127)
    lime_green = color(50, 205, 50)
    chartreuse = color(127, 255, 0)
    lawn_green = color(124, 252, 0)
    green_yellow = color(173, 255, 47)
    olive_drab = color(107, 142, 35)
    dark_olive_green = color(85, 107, 47)
    teal = color(0, 128, 128)
    dark_cyan = color(0, 139, 139)
    aquamarine = color(127, 255, 212)
    medium_aquamarine = color(102, 205, 170)
    pale_turquoise = color(175, 238, 238)
    medium_turquoise = color(72, 209, 204)
    turquoise = color(64, 224, 208)
    dark_turquoise = color(0, 206, 209)
    bright_green = color(0, 217, 58)
    emerald_green = color(80, 200, 120)

    # --- Blue Shades ---
    blue = color(0, 0, 255)
    dark_blue = color(0, 0, 139)
    medium_blue = color(0, 0, 205)
    midnight_blue = color(25, 25, 112)
    navy = color(0, 0, 128)
    royal_blue = color(65, 105, 225)
    cornflower_blue = color(100, 149, 237)
    dodger_blue = color(30, 144, 255)
    deep_sky_blue = color(0, 191, 255)
    light_sky_blue = color(135, 206, 250)
    sky_blue = color(135, 206, 235)
    steel_blue = color(70, 130, 180)
    light_steel_blue = color(176, 196, 222)
    powder_blue = color(176, 224, 230)
    cadet_blue = color(95, 158, 160)
    cyan = color(0, 255, 255)
    light_cyan = color(224, 255, 255)
    aqua = color(0, 255, 255)
    cerulean = color(0, 123, 167)
    cobalt_blue = color(0, 71, 171)
    bright_blue = color(13, 105, 172)
    ocean_blue = color(0, 119, 190)

    # --- Purple/Violet Shades ---
    purple = color(128, 0, 128)
    violet = color(238, 130, 238)
    dark_magenta = color(139, 0, 139)
    medium_orchid = color(186, 85, 211)
    orchid = color(218, 112, 214)
    plum = color(221, 160, 221)
    medium_purple = color(147, 112, 219)
    blue_violet = color(138, 43, 226)
    dark_violet = color(148, 0, 211)
    dark_orchid = color(153, 50, 204)
    rebecca_purple = color(102, 51, 153)
    indigo = color(75, 0, 130)
    amethyst = color(153, 102, 204)
    lavender = color(230, 230, 250)
    royal_purple = color(98, 37, 209)
    electric_violet = color(143, 0, 255)
    heliotrope = color(223, 115, 255)

    # --- Brown Shades ---
    brown = color(165, 42, 42)
    dark_brown = color(101, 67, 33)
    saddle_brown = color(139, 69, 19)
    sienna = color(160, 82, 45)
    chocolate = color(210, 105, 30)
    peru = color(205, 133, 63)
    rosy_brown = color(188, 143, 143)
    wheat = color(245, 222, 179)
    bisque = color(255, 228, 196)

    # --- Other Hues ---
    magenta = color(255, 0, 255)

    # --- Pastel colors ---
    pastel_red = color(255, 204, 204)
    pastel_pink = color(255, 220, 229)
    pastel_orange = color(255, 229, 204)
    pastel_yellow = color(255, 255, 204)
    pastel_green = color(204, 255, 204)
    pastel_mint = color(189, 252, 201)
    pastel_blue = color(204, 204, 255)
    pastel_cyan = color(204, 255, 255)
    pastel_lavender = color(230, 230, 250)
    pastel_purple = color(229, 204, 255)
    pastel_violet = color(200, 162, 200)
    pastel_brown = color(210, 180, 140)
    pastel_grey = color(207, 207, 207)

    # --- Vibrant/Bright colors ---
    vibrant_red = color(255, 40, 40)
    vibrant_orange = color(255, 110, 0)
    vibrant_yellow = color(255, 255, 50)
    vibrant_green = color(0, 255, 50)
    vibrant_blue = color(0, 80, 255)
    vibrant_purple = color(180, 0, 255)
    vibrant_pink = color(255, 0, 150)
    vibrant_cyan = color(0, 255, 255)
    vibrant_magenta = color(255, 0, 255)

    # --- Deep/Rich colors ---
    deep_red = color(153, 0, 0)
    deep_orange = color(204, 85, 0)
    deep_yellow = color(204, 204, 0)
    deep_green = color(0, 102, 0)
    deep_blue = color(0, 0, 153)
    deep_purple = color(102, 0, 102)
    deep_pink = color(204, 0, 102)
    deep_teal = color(0, 85, 85)
    deep_maroon = color(102, 0, 0)


from sys import platform, stdout
from functools import lru_cache
from os import environ

@lru_cache(maxsize=None)
def supports_ansi() -> bool:
    """Detects if the current terminal likely supports ANSI escape codes."""

    if environ.get('NO_COLOR'):
        return False
    if environ.get('CLICOLOR_FORCE'):
        return True

    is_tty = hasattr(stdout, 'isatty') and stdout.isatty()

    if not is_tty:
        return False

    if platform == 'win32':
        if environ.get('ANSICON') is not None or environ.get('TERM_PROGRAM') == 'vscode' or 'CONEMUANSI' in environ:
            return True
        try:
            from ctypes import windll, c_ulong, byref

            k32 = windll.kernel32
            h_out = k32.GetStdHandle(-11)
            mode = c_ulong()

            k32.GetConsoleMode(h_out, byref(mode))
            if (mode.value & 0x0004) == 0:
                k32.SetConsoleMode(h_out, mode.value | 0x0004)

            return True
        except Exception:
            pass
    else:
        term = environ.get('TERM', '').lower()
        if 'xterm' in term or 'color' in term or 'screen' in term or 'tmux' in term:
            return True

    return False


from ...core.combine_text import combine_text
from ...colors.presets import presets
from ...styles.bold import bold

_SUCCESS_COLOR = presets.bright_green

def success(*text: str) -> None:
    """Prints an success message."""

    combined_text = combine_text(*text)
    print(bold(f"{_SUCCESS_COLOR("SUCC")} {combined_text}"))
    